using System;
using System.Collections;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Threading;
using System.Text;
using System.Runtime.InteropServices;

namespace SpringClient
{
	/// <summary>
	/// Summary description for BattleManager.
	/// </summary>
	public class BattleManager
	{
		private System.Windows.Forms.ListView	battleList;
		private SortedList						m_Battles;

		private ClientMain						m_PrimaryForm;

		[DllImport(@".\Unitsync.dll")]
		private static extern int GetPrimaryModCount();
		[DllImport(@".\Unitsync.dll")]
		private static extern String GetPrimaryModName(int index);

		[DllImport(@".\Unitsync.dll")]
		private static extern Int32 GetPrimaryModIndex(string modname);
		[DllImport(@".\Unitsync.dll")]
		private static extern string GetPrimaryModArchive(Int32 index);
		[DllImport(@".\Unitsync.dll")]
		private static extern void AddAllArchives(string root);

		public BattleManager(ClientMain primaryForm, System.Windows.Forms.ListView BattleList)
		{
			//
			// TODO: Add constructor logic here
			//
			battleList = BattleList;
			m_Battles = new SortedList();
			m_PrimaryForm = primaryForm;
		}

		public void CreateBattle(BattleHostInfo bhi)
		{
			if ( m_Battles.ContainsKey(bhi.id) )
			{
				return;
			}
			
			System.Windows.Forms.ListViewItem lvi = null;
			string modname="None";
			string bname=bhi.battlename;
			if(bhi.battlename.IndexOf("|")!=-1)
			{
				modname=bhi.battlename.Substring(0,bhi.battlename.IndexOf("|"));
				bname=bhi.battlename.Substring(bhi.battlename.IndexOf("|")+1);
			}
			lvi = new System.Windows.Forms.ListViewItem(
				new string[] {bname, bhi.hostername, modname, bhi.mapname, string.Format(@"{0}\{1}", bhi.currplayers.ToString(), bhi.numplayers.ToString())});

			if ( !bhi.password )
				lvi.SubItems.Add("No");
			else
				lvi.SubItems.Add("Yes");

			if ( bhi.inProgress )
				lvi.SubItems.Add("*");
			else
				lvi.SubItems.Add("");
			lvi.SubItems.Add(bhi.id.ToString());
			bhi.lvi = lvi;
			m_Battles.Add(bhi.id, bhi);
			battleList.Items.Add(lvi);
		}

		public void MapUpdate(uint id, string mapname)
		{
			if ( !m_Battles.ContainsKey(id) )
			{
				return;
			}
			BattleHostInfo bhi = (BattleHostInfo)m_Battles[id];
			bhi.mapname = mapname;
			bhi.lvi.SubItems[3].Text = mapname;
		}

		public void SizeUpdate(uint id, byte newsize)
		{
			if ( !m_Battles.ContainsKey(id) )
			{
				return;
			}
			BattleHostInfo bhi = (BattleHostInfo)m_Battles[id];
			bhi.currplayers = newsize;
			bhi.lvi.SubItems[4].Text = string.Format(@"{0}\{1}", bhi.currplayers.ToString(), bhi.numplayers.ToString());
		}

		public void ProgressUpdate(uint id, byte newState )
		{
			if ( !m_Battles.ContainsKey(id) )
			{
				return;
			}
			BattleHostInfo bhi = (BattleHostInfo)m_Battles[id];
			bhi.inProgress = newState == 0 ? false : true;
			if ( bhi.inProgress )
				bhi.lvi.SubItems[6].Text = "*";
			else
				bhi.lvi.SubItems[6].Text = "";
		}

		public void DestroyBattle(uint id)
		{
			if ( !m_Battles.ContainsKey(id) )
			{
				return;
			}
			BattleHostInfo bhi = (BattleHostInfo)m_Battles[id];
			battleList.Items.Remove(bhi.lvi);
			m_Battles.Remove(id);
		}

		public void Clear()
		{
			battleList.Items.Clear();
			m_Battles.Clear();
		}

		public void JoinBattle(ListViewItem lvi)
		{
			try
			{
				// do nothing if a battle room is open already
				if ( Global.battleRoom != null && Global.battleRoom.Running )
					return;
				BattleLoginPrompt blp = null;
				if ( lvi.SubItems[6].Text == "*" )
				{
					MessageBox.Show("Game is in progress, cannot join");
					return;
				}

				string modname=lvi.SubItems[2].Text;
				bool haveMod=false;
				int pmCount = GetPrimaryModCount();
				for (int i = 0; i < pmCount; ++i) 
				{
					if(modname==GetPrimaryModName(i))
						haveMod=true;
				}
				if ( !haveMod )
				{
					MessageBox.Show("You dont have the gametype used in this game installed");
					return;
				}
				Global.SelectedMod = modname;
				
				if ( lvi.SubItems[5].Text == "No" )
					blp = new BattleLoginPrompt(false);
				else
					blp = new BattleLoginPrompt(true);
				if ( blp.ShowDialog() == DialogResult.Cancel )
					return;
				uint id = Convert.ToUInt32(lvi.SubItems[lvi.SubItems.Count-1].Text);
				if ( !m_Battles.ContainsKey(id) )
				{
					MessageBox.Show(string.Format("Battle id {0} not found", id), "Error");
					return;
				}
				
				BattleHostInfo bhi = (BattleHostInfo)m_Battles[id];
				if ( bhi == null )
					throw new Exception(string.Format("Battle {0} ref null", id));
				
				//IPAddress test = new IPAddress(bhi.interfaceaddress);
				//MessageBox.Show(test.ToString());
				ConnectionManager cm = new ConnectionManager();
				cm.Connect((long)bhi.interfaceaddress, (int)bhi.port);
	
				StringBuilder name = new StringBuilder(UserInfo.Name, CONSTANTS.MAXNAME);
				StringBuilder pass = new StringBuilder(blp.Password, CONSTANTS.MAXPASSWD);
				StringBuilder clan = new StringBuilder("", CONSTANTS.MAXCLAN);
				bool spec = blp.Spectator;
				while ( name.Length < CONSTANTS.MAXNAME )
					name.Append('\0');
				while ( pass.Length < CONSTANTS.MAXPASSWD )
					pass.Append('\0');
				while ( clan.Length < CONSTANTS.MAXCLAN )
					clan.Append('\0');

				byte[] buffer = new byte[4]; //+ CONSTANTS.MAXNAME + CONSTANTS.MAXPASSWD + CONSTANTS.MAXCLAN];
				BitConverter.GetBytes(UserInfo.ID).CopyTo(buffer,0);
				//Encoding.ASCII.GetBytes(pass.ToString(),0,CONSTANTS.MAXPASSWD,buffer,4);
				//Encoding.ASCII.GetBytes(clan.ToString(),0,CONSTANTS.MAXCLAN,buffer,4 + CONSTANTS.MAXPASSWD);
				//Encoding.ASCII.GetBytes(name.ToString(),0,CONSTANTS.MAXNAME,buffer,4 + CONSTANTS.MAXPASSWD + CONSTANTS.MAXCLAN);
				cm.Send(buffer);
				//cm.Send(UserInfo.ID);
				cm.Send(Encoding.ASCII.GetBytes(pass.ToString()));
				cm.Send(Encoding.ASCII.GetBytes(clan.ToString()));
				cm.Send(Encoding.ASCII.GetBytes(name.ToString()));
				if ( spec )
					cm.Send((byte)1);
				else
					cm.Send((byte)0);
				byte[] msg = new byte[1];
				cm.Read(msg);
		//		MessageBox.Show(string.Format("Read {0}", msg[0]));
				if ( (PACKETS)msg[0] == PACKETS.PACK_BATTLELOGGEDIN )
				{
					// here we are going to read in the connection id
					System.Threading.Thread.Sleep(100);
					cm.Read(msg);
		//			MessageBox.Show(string.Format("Read {0}", msg[0]));
					int intid = (int)msg[0];
					// begin the initialization
					ClientControls.UnitSync.Init(false, intid);

					// Set up the archives according to selected mod
					int idx = GetPrimaryModIndex(Global.SelectedMod);
					string baseMod = GetPrimaryModArchive(idx);
					AddAllArchives(baseMod);

					cm.Read(msg);
					byte spect = msg[0];
					bool spectator = false;
					if ( spect != 0 )
						spectator = true;
					Global.battleRoom = new BattleRoom(m_PrimaryForm, bhi.id, (int)bhi.port, false, spectator, cm, intid, 0,lvi.SubItems[2].Text);
					Global.battleRoom.Show();
				}
				else
				{
					throw new Exception("Failed to log into battle");
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			} 
		}
	}

	public class BattleHostInfo
	{
		public ListViewItem lvi;
		public uint id;
		public string battlename;
		public string clan;
		public string hostername;
		public ushort port;
		public byte numplayers;
		public uint interfaceaddress;
		public string mapname;
		public bool password;
		public byte currplayers;
		public bool inProgress;
	}
}
