using System;
using System.Collections;
using System.Threading;
using System.Windows.Forms;
using System.Net;
using System.Net.Sockets;
using System.Text;
using System.Runtime.InteropServices;
using ClientControls;
using System.IO;

using Phoenix;

namespace SpringClient
{
	/// <summary>
	/// Summary description for BattleServer.
	/// </summary>
	public class BattleServer
	{
		[DllImport("User32.dll")]
		static extern Int32 SendMessage(IntPtr hWnd, Int32 wMsg, Int32 wParam, Int32 lParam);

		[DllImport("User32.dll")]
		static extern Int32 GetScrollPos(IntPtr hWnd, Int32 nBar);

		[DllImport("User32.dll")]
		static extern Boolean GetScrollRange(IntPtr hWnd, Int32 nBar, ref Int32 lpMinPos,
			ref Int32 lpMaxPos);

		[DllImport(@".\Unitsync.dll")]
		private static extern Int32 GetPrimaryModIndex(string modname);
		[DllImport(@".\Unitsync.dll")]
		private static extern string GetPrimaryModArchive(Int32 index);
		[DllImport(@".\Unitsync.dll")]
		private static extern void AddAllArchives(string root);


		private Socket			m_Listener;

		private Hashtable		m_ConnTimers = new Hashtable();
		private Hashtable		m_CMs = new Hashtable();
		private Hashtable		m_UserInfoByConnID = new Hashtable();
		private int				m_LastID = 0;

		private Queue			m_CMPreConnects = new Queue();

		private ClientControls.ListViewEx m_lv;
		private RichTextBox		m_PrevMessages;

		private const int		MAXCONNECTIONS = 32;

		private byte			m_MaxPlayers = 0;

		private string			Password = "";
//		private StreamWriter	m_SW;
		private byte			m_UnsyncedCount = 0;
		private bool			m_bInProgress = false;

		public BattleServer(ClientControls.ListViewEx lv, RichTextBox prevMessages)
		{
			//
			// TODO: Add constructor logic here
			//

			m_lv = lv;
			m_PrevMessages = prevMessages;
		}

		public void UserDisconnected(int id )
		{
			try
			{
				if ( !m_UserInfoByConnID.ContainsKey(id))
					return;
				/*
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)m_UserInfoByConnID[id];
				if ( !bmi.m_bSpectator && Global.connMan != null && Global.connMan.Connected )
				{
					Global.connMan.Send((byte)PACKETS.PACK_BATTLESIZEUPDATE);
					Global.connMan.Send((byte)count);
				}
				*/
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)m_UserInfoByConnID[id];

				UnitSync.RemoveClient(id);
				Global.battleRoom.m_BattleStateManager.SetUnsyncAll();
				Global.battleRoom.m_BattleStateManager.SetSync(0);

				Global.battleRoom.m_BattleStateManager.RemoveMember(id);
				m_CMs.Remove(id);
				m_UserInfoByConnID.Remove(id);
				m_ConnTimers.Remove(id);
				Global.battleRoom.PopulateDisabledUnits();
				SendUserDisconnect(id);
				SendSyncedPack(0);
				SendSyncLists();
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void SendSyncLists()
		{
			try
			{
				foreach ( DictionaryEntry de in m_CMs )
				{
					ConnectionManager cm = (ConnectionManager)de.Value;
					if ( cm != null )
					{
						int id = (int)de.Key;
						string syncedlist = UnitSync.GetClientDiff(id);
						byte[] buff = new byte[syncedlist.Length+3];
						buff[0] = (byte)BATTLEPACK.PACK_SYNCLIST;
						BitConverter.GetBytes((ushort)syncedlist.Length).CopyTo(buff, 1);
						//ByteConverter.ToNetworkBytes((ushort)syncedlist.Length,buff,1);
						Encoding.ASCII.GetBytes(syncedlist, 0, syncedlist.Length, buff, 3);
						try
						{
							cm.Send(buff);
						}
						catch ( SocketException )
						{
							// we dont care about this exception, the socket will be cleaned up soon anyways
						}
						catch ( ObjectDisposedException )
						{
						}
					}
				}
			}
			catch ( ObjectDisposedException ode )
			{
				MessageBox.Show(ode.ToString());
			}
			catch ( NullReferenceException nre )
			{
				MessageBox.Show(nre.ToString());
			}
		}

		public bool isRunning()
		{
			if ( m_Listener == null || m_Listener.Connected )
				return false;
			return true;
		}

		public bool StartServer(int port, byte maxplayers)
		{
			if ( m_Listener != null )
				return false;
			try
			{
				m_bInProgress = false;
				UnitSync.Init(true, 0);

				// Set up the archives according to selected mod
				int idx = GetPrimaryModIndex(Global.SelectedMod);
				string baseMod = GetPrimaryModArchive(idx);
				AddAllArchives(baseMod);

				Global.battleRoom.m_BattleStateManager.SetSync(0);
				m_UnsyncedCount = 0;
				m_MaxPlayers = maxplayers;
				m_Listener = new Socket(AddressFamily.InterNetwork, SocketType.Stream, ProtocolType.Tcp);
				// this is needed for when you disconnect then attempt to start a new battle room
				m_Listener.SetSocketOption(SocketOptionLevel.Socket,SocketOptionName.ReuseAddress, 1);
				m_Listener.Bind(new IPEndPoint(IPAddress.Any, port));
				m_Listener.Listen(32);
				m_Listener.BeginAccept(new AsyncCallback(AcceptConnection), m_Listener);
			}
			catch ( Exception e )
			{
				MessageBox.Show(e.ToString(), "Exception in BattleServer::StartServer");
				if ( m_Listener != null )
					m_Listener.Shutdown(SocketShutdown.Both);
				return false;
			}
			return true;
		}

		public void AcceptConnection(IAsyncResult iar)
		{
			try
			{
				// blocks until a new connection is made
				Socket listener = (Socket)iar.AsyncState;
				Socket newConn = listener.EndAccept(iar);

				System.Diagnostics.Debug.WriteLine(" new connection incoming");

				// NOTE: Implement the enforcement of MAXCONNECTIONS

				// now we accept user information and if the right data is obtained, add the user to the battle room
				// and add the socket + new listener thread to the appropriate hash tables

				newConn.SetSocketOption(SocketOptionLevel.Tcp, SocketOptionName.NoDelay, (byte)1);
				ConnectionManager cm = new ConnectionManager(newConn);
				
				if ( m_bInProgress )
				{
					cm.Send((byte)PACKETS.PACK_BATTLELOGINFAILED);
					cm.Disconnect();
					listener.BeginAccept(new AsyncCallback(AcceptConnection), listener);
					return;
				}
				int messagesize = 5+CONSTANTS.MAXPASSWD+CONSTANTS.MAXCLAN+CONSTANTS.MAXNAME;

				m_CMPreConnects.Enqueue(cm);

				cm.socketHandler.Read(messagesize, new SocketHandler.delRead(HandleConnection), Global.PrimaryForm);
				// continue listening for new connections
				listener.BeginAccept(new AsyncCallback(AcceptConnection), listener);
			}
			catch ( SocketException )
			{
				// do nothing for a socket exception, since it was probably caused
				// by another thread cutting the connection
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString(), "Exception in BattleServer::WaitForConnections");
			}
		}

		public void UpdateBattleProgress ( bool running )
		{
			m_bInProgress = running;
		}

		public void SetPassword(string pass)
		{
			Password = pass;
		}

		public void HandleConnection(byte[] buffer)
		{
			try
			{
				int id = BitConverter.ToInt32(buffer, 0);
				// NOTE: THE PASSWORD FIELD IS BEING IGNORED FOR NOW
				string passwd = Encoding.ASCII.GetString(buffer, 4, CONSTANTS.MAXPASSWD).TrimEnd("\0".ToCharArray());
				string clan = Encoding.ASCII.GetString(buffer, 4+CONSTANTS.MAXPASSWD, CONSTANTS.MAXCLAN).TrimEnd("\0".ToCharArray());
				string name = Encoding.ASCII.GetString(buffer, 4+CONSTANTS.MAXPASSWD+CONSTANTS.MAXCLAN, CONSTANTS.MAXNAME).TrimEnd("\0".ToCharArray());
				byte spec = buffer[buffer.Length-1];
				ConnectionManager cm = (ConnectionManager)m_CMPreConnects.Dequeue();
				
				IDictionaryEnumerator ide = m_UserInfoByConnID.GetEnumerator();
				int count = 1;
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo currbmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					if ( !currbmi.m_bSpectator )
						count++;
				}

				// enforce max players count
				if ( count >= m_MaxPlayers && spec == 0 )
				{
					cm.Send((byte)PACKETS.PACK_BATTLELOGINFAILED);
					cm.Disconnect();
					return;
				}

				if ( Password.Length != 0 )
				{
					if ( passwd != Password )
					{
						cm.Send((byte)PACKETS.PACK_BATTLELOGINFAILED);
						cm.Disconnect();
						return;
					}
				}
				// so we will assume the connection succeeded and handle multiple cases later
				m_LastID++;

				while ( m_CMs.Contains(m_LastID) )
				{
					if ( m_LastID > 10000 )
						m_LastID = 1;
					else
						m_LastID++;
				}

				BattleStateManager.BattleMemberInfo bmi = null;

				if ( spec == 0 )
					bmi = Global.battleRoom.AddMember(false, m_LastID, false, (BattleStateManager.TeamColor)0, -1, name, clan, 0);
				else
					bmi = Global.battleRoom.AddMember(true, m_LastID, false, (BattleStateManager.TeamColor)0, -1, name, clan, 0);
				
				// once someone joins the room, everyone is unsynced
				Global.battleRoom.m_BattleStateManager.SetUnsyncAll();
				m_UnsyncedCount++;

				bmi.m_ConnID = m_LastID;
				m_UserInfoByConnID.Add(bmi.m_ConnID, bmi);

				m_CMs.Add(bmi.m_ConnID, cm);

				System.Diagnostics.Debug.WriteLine(" new connection complete");

				cm.socketHandler.SetID(bmi.m_ConnID);
				
				cm.Send((byte)PACKETS.PACK_BATTLELOGGEDIN);
				
				cm.Send((byte)m_LastID);
				cm.Send(spec);
				
				SendNewUser(bmi);

				// now we send a battle list
				SendBattleList(cm,1);

				//team should be decided on the server not the client to avoid confusion when several ppl join at once
				if (spec==0)
				{
					foreach ( BattleStateManager.TeamInfo ti in Global.battleRoom.m_BattleStateManager.Teams )
					{
						if ( ti.m_Members.Count == 0 )
						{
							Global.battleRoom.ChangeTeam(bmi.m_ID, (BattleStateManager.TeamColor)ti.m_TeamID);

							//change team msg isnt automatically sent to the joiner by the above
							cm.Send((byte)BATTLEPACK.PACK_CHANGETEAM);
							cm.Send(BitConverter.GetBytes(bmi.m_ID));
							cm.Send((byte)ti.m_TeamID);
							break;
						}
					}
				}

				// and send mapname
				cm.Send((byte)BATTLEPACK.PACK_MAP);
				cm.Send(Encoding.ASCII.GetBytes(Global.battleRoom.SMapName.PadRight(CONSTANTS.MAXMAPNAME, '\0')));
				
				//send resources etc
				BattleStateManager bsm=Global.battleRoom.m_BattleStateManager;
				SendChangeResourcesMessage(0,bsm.metal,bsm.energy);
				SendChangeMaxUnitsMessage(0,bsm.maxUnits);
				SendChangeStartPosTypeMessage(0,bsm.startPosType);
				SendChangeGameGoalTypeMessage(0,bsm.gameGoalType);

				// and fire the main message loop
				
				ConnectionTimer connTimer = new ConnectionTimer(this, cm, bmi.m_ID);
				m_ConnTimers.Add(bmi.m_ConnID, connTimer);

				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void SendNewUser(BattleStateManager.BattleMemberInfo bmi)
		{
			try
			{
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
				int count = 1;
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi2 = (BattleStateManager.BattleMemberInfo)ide.Value;
					if ( !bmi2.m_bSpectator )
						count++;
					// don't send the message to the person who made the change
					if ( bmi2.m_ConnID == bmi.m_ConnID )
						continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_NEWUSER);
						//cm.Send(bmi.m_ConnID);
						cm.Send(BitConverter.GetBytes(bmi.m_ConnID));
						cm.Send(Encoding.ASCII.GetBytes(bmi.m_Clan.PadRight(CONSTANTS.MAXCLAN, '\0')));
						cm.Send(Encoding.ASCII.GetBytes(bmi.m_Name.PadRight(CONSTANTS.MAXNAME, '\0')));
						cm.Send((byte)bmi.m_TeamID);
						cm.Send((byte)bmi.m_InternalID);
						if ( bmi.m_bSpectator )
							cm.Send((byte)1);
						else
							cm.Send((byte)0);
						//	spectator, id, self, teamid, internalteamid, name, clan, ready
					}
				}
				if ( !bmi.m_bSpectator && Global.connMan != null && Global.connMan.Connected )
				{
					Global.connMan.Send((byte)PACKETS.PACK_BATTLESIZEUPDATE);
					Global.connMan.Send((byte)count);
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void SendBattleList(ConnectionManager cm,byte initialize)
		{
	//		try
			{

				cm.Send((byte)BATTLEPACK.PACK_LIST);

				cm.Send((byte)initialize);

				cm.Send((byte)Global.battleRoom.m_BattleStateManager.AllyTeams.Length);
				IEnumerator ie = Global.battleRoom.m_BattleStateManager.AllyTeams.GetEnumerator();
				while ( ie.MoveNext() )
				{
					BattleStateManager.AllyTeamInfo ati = (BattleStateManager.AllyTeamInfo)ie.Current;
					cm.Send((byte)ati.m_AlliedTo.Count );
					foreach( byte to in ati.m_AlliedTo )
						cm.Send(to );
				}

				cm.Send((byte)Global.battleRoom.m_BattleStateManager.Teams.Length);
				ie = Global.battleRoom.m_BattleStateManager.Teams.GetEnumerator();
				while ( ie.MoveNext() )
				{
					BattleStateManager.TeamInfo ti = (BattleStateManager.TeamInfo)ie.Current;
					cm.Send((byte)ti.m_AllyTeamID);
					cm.Send((byte)ti.m_InternalID);
					cm.Send((byte)ti.m_Side);
				}

				cm.Send((byte)Global.battleRoom.m_BattleStateManager.OrderedMembers.Count);
				ie = Global.battleRoom.m_BattleStateManager.OrderedMembers.GetEnumerator();
				//	IDictionaryEnumerator ide = Global.battleRoom.m_BattleStateManager.Members.GetEnumerator();
				while ( ie.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ie.Current;
					StringBuilder sb = new StringBuilder(bmi.m_Clan);
					StringBuilder sb2 = new StringBuilder(bmi.m_Name);
					while ( sb.Length < CONSTANTS.MAXCLAN )
						sb.Append('\0');
					while ( sb2.Length < CONSTANTS.MAXNAME )
						sb2.Append('\0');			
					cm.Send(BitConverter.GetBytes(bmi.m_ID));
					cm.Send(Encoding.ASCII.GetBytes(sb.ToString()));
					cm.Send(Encoding.ASCII.GetBytes(sb2.ToString()));
					cm.Send((byte)bmi.m_TeamID);
					cm.Send(bmi.m_InternalID);
					if ( !bmi.m_bSpectator &&  bmi.m_checkReady.Checked )
						cm.Send((byte)1);
					else
						cm.Send((byte)0);
					if ( bmi.m_bSpectator )
						cm.Send((byte)1);
					else
						cm.Send((byte)0);
				}
			}
		}

		public void SendChangeTeamMessage(int id, BattleStateManager.TeamColor to)
		{
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					// don't send the message to the person who made the change
					if ( bmi.m_ID == id )
						continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGETEAM);
						cm.Send(BitConverter.GetBytes(id));
						cm.Send((byte)to);
					}
				}
		}

		public void SendStartBattle()
		{
			try
			{
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						//SendBattleList(cm,0);	//make sure to sync everyones battle list before start
						cm.Send((byte)BATTLEPACK.PACK_START);
					}
				}
//				Global.battleRoom.ChangeReady(true, 0, false);
				BattleStateManager.BattleMemberInfo m_Self = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[0];
				m_Self.m_checkReady.Checked = false;
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void SendChangeAllyTeamMessage(byte team, byte to)
		{
	//			MessageBox.Show("SEND ALLY TEAM MESSAGE");
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
	//			BattleStateManager.BattleMemberInfo bmi2 = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[sender];
	//			byte team = (byte)bmi2.m_TeamID;
				while ( ide.MoveNext() )
				{
	//				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					// don't send the message to the person who made the change
	//				if ( bmi.m_ID == sender )
	//					continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGEALLYTEAM);
						cm.Send(team);//Global.battleRoom.m_BattleStateManager.Teams[(byte)bmi.m_TeamID].m_AllyTeamID);
						cm.Send((byte)to);
					}
				}
		}

		public void SendChangeResourcesMessage(int sender, int metal,int energy)
		{
				//MessageBox.Show("SEND Resource MESSAGE");
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					// don't send the message to the person who made the change
					if ( bmi.m_ID == sender )
						continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGERESOURCES);
						cm.Send(BitConverter.GetBytes(metal));
						cm.Send(BitConverter.GetBytes(energy));
					}
				}
		}

		public void SendChangeMaxUnitsMessage(int sender, int maxUnits)
		{
				//MessageBox.Show("SEND MaxUnits MESSAGE");
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					// don't send the message to the person who made the change
					if ( bmi.m_ID == sender )
						continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGEMAXUNITS);
						cm.Send(BitConverter.GetBytes(maxUnits));
					}
				}
		}

		public void SendChangeStartPosTypeMessage(int sender, int type)
		{
				//MessageBox.Show("SEND MaxUnits MESSAGE");
				IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
					// don't send the message to the person who made the change
					if ( bmi.m_ID == sender )
						continue;
					ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGESTARTPOSTYPE);
						cm.Send(BitConverter.GetBytes(type));
					}
				}
		}

		public void SendChangeGameGoalTypeMessage(int sender, int type)
		{
			//MessageBox.Show("SEND MaxUnits MESSAGE");
			IDictionaryEnumerator ide = this.m_UserInfoByConnID.GetEnumerator();
			while ( ide.MoveNext() )
			{
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide.Value;
				// don't send the message to the person who made the change
				if ( bmi.m_ID == sender )
					continue;
				ConnectionManager cm = (ConnectionManager)m_CMs[(int)ide.Key];
				lock ( cm )
				{
					cm.Send((byte)BATTLEPACK.PACK_CHANGEGAMEGOALTYPE);
					cm.Send(BitConverter.GetBytes(type));
				}
			}
		}

		public void MessageLoop(byte[] buffer, int currID)
		{
			ConnectionManager cm = null;

			try
			{
				cm = (ConnectionManager)m_CMs[currID];
				if ( cm == null )
					return;
			
				switch ( (BATTLEPACK)buffer[0] )
				{
					case BATTLEPACK.PACK_MESSAGE:
						cm.socketHandler.Read(2, new SocketHandler.delReadID(HandlePackBattleMessage), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_LEAVE:
						HandlePackLeaveBattle((BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[currID]);
						cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_CHANGETEAM:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackChangeTeam), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_CHANGEALLYTEAM:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackChangeAllyTeam), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_READY:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackReady), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_KEEPALIVE:
						// keep alives should result in message loop being fired again
						cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_ALLIANCE:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackAddAlliance), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_REMOVEALLIANCE:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackRemoveAlliance), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_CHANGESIDE:
						cm.socketHandler.Read(1, new SocketHandler.delReadID(HandlePackChangeSide), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_SYNCLIST:
						cm.socketHandler.Read(2, new SocketHandler.delReadID(HandlePackSyncList), Global.PrimaryForm);
						break;
					case BATTLEPACK.PACK_SYNCED:
						HandlePackSynced(currID);
						break;
					default:
						throw new Exception(string.Format("Bad message from {0}", ((BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[currID]).m_Name));
				}
			}
			catch ( ObjectDisposedException )
			{
				// do nothing, this means the connectino was lost
				UserDisconnected(currID);
			}
			catch ( SocketException )
			{
				// do nothing, this means the connection was lost
				UserDisconnected(currID);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString(), "Exception in BattleServer::ConnectionHandler()");
			}
		}

		private void HandlePackSyncList( byte[] buffer, int id )
		{
			try
			{
				ushort len = BitConverter.ToUInt16(buffer, 0);
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				try
				{
					cm.socketHandler.Read(len, new SocketHandler.delReadID(HandlePackSyncListHelper), Global.PrimaryForm);
				}
				catch ( ObjectDisposedException ode )
				{
					MessageBox.Show(ode.ToString());
				}
				catch ( SocketException )
				{
					cm.Disconnect();
				}
			}
			catch ( ArgumentOutOfRangeException are )
			{
				// these are temporary
				MessageBox.Show(are.ToString());
			}
		}

		private void HandlePackSyncListHelper( byte[] buffer, int id )
		{
			try
			{
				string list = Encoding.ASCII.GetString(buffer);
				while ( !UnitSync.UnitsProcessed() )
				{
					MessageBox.Show("Waiting on unit syncer to finish processing units");
					Thread.Sleep(TimeSpan.FromSeconds(1.0));
				}
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				if ( !UnitSync.AddClient(id, list) )
				{
					MessageBox.Show("Failed to add client to unit sync server, kicking");
					cm.Disconnect();
					return;
				}
				m_UnsyncedCount--;
				if ( m_UnsyncedCount == 0 )
				{
					Global.battleRoom.m_BattleStateManager.SetSync(0);
					SendSyncedPack(0);
					Global.battleRoom.PopulateDisabledUnits();
				}
				SendSyncLists();
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( ObjectDisposedException ode )
			{
				MessageBox.Show(ode.ToString());
			}
			catch ( NullReferenceException nre )
			{
				MessageBox.Show(nre.ToString());
			}
		}

		private void SendSyncedPack( int id )
		{
			try
			{
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				byte[] buf = new byte[5];
				while ( ide.MoveNext() )
				{
					if ( (int)ide.Key == id )
						continue;
					ConnectionManager cm = (ConnectionManager)ide.Value;
					buf[0] = (byte)BATTLEPACK.PACK_SYNCED;
					BitConverter.GetBytes(id).CopyTo(buf, 1);
					//ByteConverter.ToNetworkBytes(id,buf,1);
					try
					{
						cm.Send(buf);
					}
					catch ( SocketException )
					{
						cm.Disconnect();
					}
				}
			}
			catch ( NullReferenceException nre )
			{
				MessageBox.Show(nre.ToString());
			}
			catch ( InvalidOperationException ioe )
			{
				MessageBox.Show(ioe.ToString());
			}
		}

		private void HandlePackSynced ( int id )
		{
			Global.battleRoom.m_BattleStateManager.SetSync(id);
			SendSyncedPack(id);
			ConnectionManager cm = (ConnectionManager)m_CMs[id];
			cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
		}

		private void HandlePackAddAlliance( byte[] buffer, int id )
		{
			try
			{
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[id];
				byte team = (byte)bmi.m_TeamID;
				byte allyteam = Global.battleRoom.m_BattleStateManager.Teams[team].m_AllyTeamID;
				byte to = buffer[0];
				if ( bmi.m_InternalID == 0 && Global.battleRoom.m_BattleStateManager.Teams[team].m_InternalID == 0 )
				{	
					Global.battleRoom.AddAlliance(allyteam, to);
					SendAlliance(allyteam, to);
				}
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackRemoveAlliance( byte[] buffer, int id )
		{
			try
			{
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[id];
				byte team = (byte)bmi.m_TeamID;
				byte allyteam = Global.battleRoom.m_BattleStateManager.Teams[team].m_AllyTeamID;
				byte to = buffer[0];
				if ( bmi.m_InternalID == 0 && Global.battleRoom.m_BattleStateManager.Teams[team].m_InternalID == 0 )
				{
					Global.battleRoom.RemoveAlliance(allyteam, to);
					SendRemoveAlliance(allyteam, to);
				}
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackChangeSide(byte[] buffer, int id )
		{
			try
			{
				BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[id];
				byte team = (byte)bmi.m_TeamID;
				byte to = buffer[0];
				if ( bmi.m_InternalID == 0 )
				{
					Global.battleRoom.Changeside(team, to);
					SendChangeSide(team, to);
				}
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackReady(byte[] buffer, int id )
		{
			try
			{
				if ( buffer[0] == 0 )
					Global.battleRoom.ChangeReady(false, id, false);
				else
					Global.battleRoom.ChangeReady(false, id, true );
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		// expects one byte
		private void HandlePackChangeTeam(byte[] buffer, int id)//Socket s, BattleStateManager.BattleMemberInfo ui)
		{	
			try
			{
				BattleStateManager.TeamColor tc = (BattleStateManager.TeamColor)buffer[0];
				BattleStateManager.BattleMemberInfo ui = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[id];

				Global.battleRoom.ChangeTeam(ui.m_ID, tc);
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);

			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		// expects one byte
		private void HandlePackChangeAllyTeam(byte[] buffer, int id)//Socket s, BattleStateManager.BattleMemberInfo ui)
		{
			try
			{
				BattleStateManager.TeamColor tc = (BattleStateManager.TeamColor)buffer[0];
				BattleStateManager.BattleMemberInfo ui = (BattleStateManager.BattleMemberInfo)Global.battleRoom.m_BattleStateManager.Members[id];

				if ( ui.m_InternalID == 0 )
				{
					Global.battleRoom.ChangeAllyTeam(false, (byte)ui.m_TeamID, (byte)tc);
				}
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}
		
		// expect 2 bytes
		private void HandlePackBattleMessage(byte[] buffer, int id)
		{
			try
			{
				ushort len = BitConverter.ToUInt16(buffer, 0);
				int messagesize = 4+(int)len;
;
				ConnectionManager cm = (ConnectionManager)m_CMs[id];
				
				if ( len > CONSTANTS.MAXMESSAGELEN )
				{
					cm.Disconnect();
				}
				else
				{
					cm.socketHandler.Read(messagesize, new SocketHandler.delRead(HandlePackBattleMessageQuote), Global.PrimaryForm);
					cm.socketHandler.Read(1, new SocketHandler.delReadID(MessageLoop), Global.PrimaryForm);
				}
			}
			catch ( Exception ex)
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void HandlePackBattleMessageQuote(byte[] buffer)
		{
			try
			{
				int id = BitConverter.ToInt32(buffer, 0);
				string quote = Encoding.ASCII.GetString(buffer, 4, buffer.Length-4);
				SendMessage(id, quote);
				Global.battleRoom.PostMessage(id, quote);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		private void SendUserDisconnect(int id)
		{
				IDictionaryEnumerator ide2 = m_UserInfoByConnID.GetEnumerator();
				byte count = 0;
				while ( ide2.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi = (BattleStateManager.BattleMemberInfo)ide2.Value;
					if ( !bmi.m_bSpectator )
						count++;
				}
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					// host has id of 0
					if ( (int)ide.Key == 0 )
						continue;
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_LEAVE);
						cm.Send(BitConverter.GetBytes(id));
					}
				}
				if ( Global.connMan != null && Global.connMan.Connected )
				{
					Global.connMan.Send((byte)PACKETS.PACK_BATTLESIZEUPDATE);
					Global.connMan.Send(count);
				}
		}

		private void HandlePackLeaveBattle(BattleStateManager.BattleMemberInfo bmi)
		{
			try
			{
		//		Global.battleRoom.m_BattleStateManager.RemoveMember(bmi.m_ConnID);
				
				UserDisconnected(bmi.m_ConnID);

				// here is where we send a leave message to everyone but the person who is leaving
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				byte count = 1;
				
				while ( ide.MoveNext() )
				{
					BattleStateManager.BattleMemberInfo bmi2 = (BattleStateManager.BattleMemberInfo)m_UserInfoByConnID[(int)ide.Key];
					// don't send a message to the leaving user
					if ( object.ReferenceEquals(bmi2, bmi) )
						continue;
					if ( !bmi2.m_bSpectator )
						count++;
				}

				if ( !bmi.m_bSpectator && Global.connMan != null && Global.connMan.Connected )
				{
					Global.connMan.Send((byte)PACKETS.PACK_BATTLESIZEUPDATE);
					Global.connMan.Send((byte)count);
				}
				
				/*
				ConnectionManager cm2 = (ConnectionManager)m_CMs[bmi.m_ID];
				cm2.Disconnect();
				m_CMs.Remove(bmi.m_ID);
				m_ConnTimers.Remove(bmi.m_ID);
				m_UserInfoByConnID.Remove(bmi.m_ID);
				*/
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void BootUser(BattleStateManager.BattleMemberInfo bmi)
		{
			try
			{
				ConnectionManager cm = (ConnectionManager)m_CMs[bmi.m_ConnID];
				cm.Send((byte)BATTLEPACK.PACK_BOOTED);
			//	cm.Disconnect();
			//	UserDisconnected(bmi.m_ConnID);
			//	Global.battleRoom.m_BattleStateManager.RemoveMember(bmi.m_ConnID);
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public bool SendMessage(int id, string data)
		{
				if ( data.Length >= 65536 )
					return false; // too long a message
				ushort len = (ushort)data.Length;
				byte[] msg = Encoding.ASCII.GetBytes(data);
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					// host has id of 0
					if ( (int)ide.Key == 0 )
						continue;
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_MESSAGE);
						cm.Send(BitConverter.GetBytes(len));
						cm.Send(BitConverter.GetBytes(id));
						cm.Send(msg);
					}
				}
			return true;
		}

		public void SendChangeMap( string filename )
		{
			IDictionaryEnumerator ide = m_CMs.GetEnumerator();
			while ( ide.MoveNext() )
			{
				// host has id of 0
				if ( (int)ide.Key == 0 )
					continue;
				ConnectionManager cm = (ConnectionManager)ide.Value;
				lock ( cm )
				{
					cm.Send((byte)BATTLEPACK.PACK_MAP);
					cm.Send(Encoding.ASCII.GetBytes(filename.PadRight(CONSTANTS.MAXMAPNAME, '\0')));
				}
			}
			if ( Global.connMan != null && Global.connMan.Connected )
			{
				Global.connMan.Send((byte)PACKETS.PACK_MAPUPDATE);
				Global.connMan.Send(Encoding.ASCII.GetBytes(filename.PadRight(CONSTANTS.MAXMAPNAME, '\0')));
			}
		}

		public void SendChangeSide( byte team, byte to)
		{
			try
			{
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_CHANGESIDE);
						cm.Send(team);
						cm.Send(to);
					}
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void SendAlliance( byte team, byte to )
		{
			try
			{
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_ALLIANCE);
						cm.Send(team);
						cm.Send(to);
					}
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void SendRemoveAlliance( byte team, byte to )
		{
			try
			{
				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					ConnectionManager cm = (ConnectionManager)ide.Value;
					lock ( cm )
					{
						cm.Send((byte)BATTLEPACK.PACK_REMOVEALLIANCE);
						cm.Send(team);
						cm.Send(to);
					}
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
		}

		public void SendReadyState(int id, bool state)
		{
			IDictionaryEnumerator ide = m_CMs.GetEnumerator();
			while ( ide.MoveNext() )
			{
				// don't send back to the sender
				if ( (int)ide.Key == id )
					continue;
				ConnectionManager cm = (ConnectionManager)ide.Value;
				lock ( cm )
				{
					cm.Send((byte)BATTLEPACK.PACK_READY);
					cm.Send(BitConverter.GetBytes(id));
					if ( state )
						cm.Send((byte)1);
					else
						cm.Send((byte)0);
				}
			}
		}

		public bool PostMessage(int userid, string data)
		{
			if ( Global.battleRoom.m_BattleStateManager.Members.ContainsKey(userid)  )
			{
				BattleStateManager.BattleMemberInfo ui = (BattleStateManager.BattleMemberInfo)
					Global.battleRoom.m_BattleStateManager.Members[userid];
				m_PrevMessages.AppendText(ui.m_Name.ToString() + ": ");
			}
			else
				m_PrevMessages.AppendText(userid.ToString() + ": ");
		
			m_PrevMessages.AppendText(data + "\n");
			m_PrevMessages.Update();
			if ( !m_PrevMessages.Focused )
				SendMessage(m_PrevMessages.Handle, Win32.WM_VSCROLL, Win32.SB_BOTTOM, 0);
			return true;
		}

		public void Stop()
		{
			try
			{
				UnitSync.Free();
				Global.battleRoom.m_BattleStateManager.Clear(true);
				try
				{
					if ( m_Listener != null )
					{
						m_Listener.Shutdown(SocketShutdown.Both);
						m_Listener.Close();
					}
					m_Listener = null;
				}
				catch ( SocketException )
				{
				}

				IDictionaryEnumerator ide = m_CMs.GetEnumerator();
				while ( ide.MoveNext() )
				{
					try
					{
						ConnectionManager cm = (ConnectionManager)ide.Value;
						cm.Disconnect();
					}
					catch ( SocketException )
					{
					}
				}
			}
			catch ( Exception ex )
			{
				MessageBox.Show(ex.ToString());
			}
			finally
			{
				m_CMs.Clear();
				m_UserInfoByConnID.Clear();
				m_ConnTimers.Clear();
				Global.connMan.Send((byte)PACKETS.PACK_DESTROYBATTLE);
			}
		}
	}
}
