using System;
using System.Windows.Forms;
using System.Net.Sockets;

namespace SpringClient
{
	/// <summary>
	/// This checks the status of connections in the battle room periodically since
	/// all calls are asynchrous and we won't know when a connection is lost
	/// </summary>
	public class ConnectionTimer
	{
		private ConnectionManager m_CM;
		private int				  m_ID;
		private BattleServer	  m_Server;
		private System.Threading.Timer  m_Timer;
		public ConnectionTimer(BattleServer bs, ConnectionManager cm, int id)
		{
			//
			// TODO: Add constructor logic here
			//
			m_Server = bs;
			m_CM = cm;
			m_ID = id;

			m_Timer = new System.Threading.Timer(new System.Threading.TimerCallback(timerCallback), m_Timer, 4000, 4000);
		}

		private void timerCallback( object state )
		{
			try
			{
				lock ( m_CM )
				{
					m_CM.Send((byte)BATTLEPACK.PACK_KEEPALIVE);
					if ( !m_CM.Connected )
					{
						m_Server.UserDisconnected(m_ID);
						System.Diagnostics.Debug.WriteLine(" disconnected ");
						m_Timer.Dispose();
						m_Timer = null;
					}
					else
						System.Diagnostics.Debug.WriteLine(" still connected ");
				}
			}
			catch ( System.Net.Sockets.SocketException )
			{
				m_Server.UserDisconnected(m_ID);
				m_Timer.Dispose();
				m_Timer = null;
			}
			catch ( Exception ex )
			{
				m_Server.UserDisconnected(m_ID);
				m_Timer.Dispose();
				m_Timer = null;
				MessageBox.Show(ex.ToString());
			}
		}

		public void Stop()
		{
			if ( m_Timer != null )
			{
				m_Timer.Dispose();
				m_Timer = null;
			}
		}
		/*
				private void ConnectionTimer_Tick(object sender, EventArgs e)
				{
					try
					{
						m_CM.Send((byte)BATTLEPACK.PACK_KEEPALIVE);
						if ( !m_CM.Connected )
						{
							m_Server.UserDisconnected(m_ID);
							System.Diagnostics.Debug.WriteLine(" disconnected ");
							Stop();
						}
						else
							System.Diagnostics.Debug.WriteLine(" still connected ");
					}
					catch ( SocketException )
					{
						m_Server.UserDisconnected(m_ID);
						Stop();
					}
					catch ( Exception ex )
					{
						m_Server.UserDisconnected(m_ID);
						Stop();
						MessageBox.Show(ex.ToString());
					}
				}
				*/
	}
}
